/***************************************************************************
 *
 * Content      : Implementation of the Linux interface
 *
 * COPYRIGHT (c) 2004  Matrox Electronic Systems Ltd.
 * All Rights Reserved
 *
 ****************************************************************************/

#include <linux/module.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/highmem.h>
#include <linux/interrupt.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <asm/uaccess.h>
#include <linux/cdev.h>
#include <linux/kthread.h>
#include <linux/cpufreq.h>
#include <asm/mtrr.h>
#include <linux/completion.h>
#include <linux/tty.h>
#include <linux/serial_core.h>
#include <linux/semaphore.h>


#include "linux-interface.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#error You are using an unsupported Legacy kernel. It is time to UPDATE YOUR KERNEL.
#endif

int MDrv_inatomic(void)
{
   return in_atomic();
}

int MDrv_irqsdisabled(void)
{
   return irqs_disabled();
}

void MDrv_localirqsave(unsigned long *flag)
{
   local_irq_save(*flag);
}

void MDrv_localirqrestore(unsigned long flag)
{
   local_irq_restore(flag);
}

void *MDrv_memcpy (void *dest, const void *src, int n)
{
   BUG_ON (dest == NULL);
   return memcpy (dest, src, n);
}

int MDrv_memcpy_s (void *dest, unsigned int size, const void *src, unsigned int count)
{
   BUG_ON (dest == NULL);
   if (size < count)
      return -1;

   memcpy (dest, src, count);
   return 0;
}

void *MDrv_memmove (void *dest, const void *src, int n)
{
   BUG_ON (dest == NULL);
   BUG_ON (src == NULL);
   return memmove (dest, src, n);
}

int MDrv_memmove_s (void *dest, unsigned int size, const void *src, unsigned int count)
{
   BUG_ON (dest == NULL);
   BUG_ON (src == NULL);
   if (size < count)
      return -1;

   memmove (dest, src, count);
   return 0;
}

void *MDrv_memset (void *str, int c, int n)
{
   BUG_ON (str == NULL);
   return memset (str, c, n);
}

void *MDrv_memset16 (void *dest, int patern, int count)
{
   BUG_ON (dest == NULL);
   __asm__ __volatile__ (
      "rep \n\t"
      "stosw \n\t"
      :
      : "a" (patern), "c" (count), "D" (dest)
      : "memory"
      );

   return dest;
}

void *MDrv_memset32 (void *dest, int patern, int count)
{
   BUG_ON (dest == NULL);
   __asm__ __volatile__ (
      "rep \n\t"
      "stosl \n\t"
      :
      : "a" (patern), "c" (count), "D" (dest)
      : "memory"
      );

   return dest;
}

int MDrv_memcmp(const void *s1, const void *s2, int n)
{
   return memcmp (s1, s2, n);
}

int MOs_strlen (const char *str)
{
   BUG_ON (str == NULL);
   return strlen (str);
}

int MOs_strnlen (const char *str, unsigned int size)
{
   BUG_ON (str == NULL);
   return strnlen (str, size);
}

char *MOs_strcpy(char *dest, const char *src)
{
   BUG_ON (dest == NULL);
   return strcpy (dest, src);
}

int MOs_strcpy_s (char *dest, unsigned int size, const char *src)
{
   BUG_ON (dest == NULL);
   strncpy (dest, src, size);
   dest[size-1] = '\0';
   return (((strnlen(src, size) + 1) > size) ? -1 : 0);
}

char *MOs_strncpy(char *dest, const char *src, unsigned int size)
{
   BUG_ON (dest == NULL);
   return strncpy (dest, src, size);
}

int MOs_strncpy_s (char *dest, unsigned int size, const char *src, unsigned int count)
{
   if ((strnlen(src, count) + 1) > size)
   {
      dest[0] = '\0';
      return -1;
   }
   strncpy (dest, src, count);
   dest[size-1] = '\0';
   return 0;
}

char *MOs_strcat(char *dest, const char *src)
{
   BUG_ON (dest == NULL);
   return strcat (dest, src);
}

int MOs_strcat_s (char *dest, unsigned int size, const char *src)
{
   int ret_val = 0;
   BUG_ON (dest == NULL);
   ret_val = (((strlen(dest) + strlen(src) + 1) > size) ? -1 : 0);
   strncat (dest, src, size);
   dest[size-1] = '\0';
   return ret_val;
}

char *MOs_strncat(char *dest, const char *src, unsigned int size)
{
   BUG_ON (dest == NULL);
   return strncat (dest, src, size);
}

int MOs_strncat_s (char *dest, unsigned int size, const char *src, unsigned int count)
{
   if ((strlen(dest) + strnlen(src, count) + 1) > size)
   {
      dest[0] = '\0';
      return -1;
   }
   strncat (dest, src, count);
   dest[size-1] = '\0';
   return 0;
}

int MOs_strcmp(const char *s1, const char *s2)
{
   return strcmp (s1, s2);
}

int MOs_strncmp(const char *s1, const char *s2, int n)
{
   return strncmp (s1, s2, n);
}

int MOs_strtolower (unsigned char *Buffer)
{
   unsigned char *c = NULL;
   for (c = Buffer; *c; ++c)
      *c = tolower(*c);
   return 0;
}

int MDrv_print (const char *format, ...)
{
   va_list arg_lst;
   int ret = 0;

   va_start(arg_lst, format);
   ret = vprintk(format, arg_lst);
   va_end(arg_lst);

   return ret;
}

int MOs_sprintf (char *str, const char *format, ...)
{
   int ret = 0;
   va_list arg_lst;
   BUG_ON (str == NULL);

   va_start(arg_lst, format);
   ret = vsprintf(str, format, arg_lst);
   va_end(arg_lst);

   return ret;
}

int MOs_vsprintf_s (char *str, int size, const char *format, va_list list)
{
   return vsprintf(str, format, list);
}

int MDrv_snprintf (char *str, int size, const char *format, ...)
{
   int ret = 0;
   va_list arg_lst;
   BUG_ON (str == NULL);

   va_start(arg_lst, format);
   ret = vsnprintf(str, size, format, arg_lst);
   va_end(arg_lst);

   return ret;
}

int MDrv_sscanf (const char *str, const char *format, ...)
{
   int ret = 0;
   va_list arg_lst;
   BUG_ON (str == NULL);

   va_start(arg_lst, format);
   ret = vsscanf(str, format, arg_lst);
   va_end(arg_lst);

   return ret;
}

void MDrv_atomic_test(void)
{
   if (in_interrupt())
      MDrv_print("In Interrupt\n");
   if (in_atomic())
      MDrv_print("In Atomic\n");
   if (in_irq())
      MDrv_print("In IRQ\n");
   if (irqs_disabled())
      MDrv_print("IRQ disabled\n");
}

void *MDrv_memalloc (int size)
{
   void *temp = NULL;
   if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
   {
      temp = kmalloc (size, GFP_ATOMIC);
   }
   else
   {
      temp = kmalloc (size, GFP_KERNEL);
   }
   if (!temp)
   {
      printk("Allocation failled for %d bytes\n", size);
      temp = NULL;
   }
   if (IS_ERR(temp))
   {
      printk("Allocation Error for %d bytes\n", size);
      temp = NULL;
   }

   return temp;
}

void MDrv_memfree (const void *p)
{
   kfree (p);
}

void *MDrv_vmemalloc (int size)
{
   BUG_ON (in_interrupt() || in_atomic());
   return vmalloc (size);
}

void MDrv_vmemfree (void *p)
{
   vfree (p);
}

unsigned long MDrv_allocate_stack (void **StackPtr, unsigned long StackSize)
{
   // Since we're running with interupts disabled, does this realy matter?

   // Ignore Stack size: use native
   // We need to allocate a THREAD_SIZE aligned stack so we allocate 2x STACK_SIZE plus space to save the base address
   u8 *slab = MDrv_memalloc (2UL* THREAD_SIZE + sizeof(void*));
   if (slab)
   {
      struct thread_info *thread = NULL;
      void **newStack = (void**)(((unsigned long)slab + THREAD_SIZE) & (~(unsigned long)(THREAD_SIZE-1)));
      *(newStack + (THREAD_SIZE/sizeof(void*))) = slab;
      *StackPtr = newStack;

      //Initialise the stack
      thread = (struct thread_info *)newStack;
      memset (thread, 0, sizeof(struct thread_info));
      thread->preempt_count = HARDIRQ_OFFSET;

      return THREAD_SIZE;
   }
   *StackPtr = NULL;
   return 0;
}

void MDrv_free_stack (void *StackPtr)
{
   MDrv_memfree(*(void**)(((u8 *)StackPtr) + THREAD_SIZE));
}

void *MDrv_fopen (const char *filename, int attr)
{
   struct file * ret = NULL;
   BUG_ON (filename == NULL);

   if (attr == MDRV_FRDONLY)
      ret = filp_open (filename, O_RDONLY, 0);

   if ((ret == NULL) || IS_ERR(ret))
      return NULL;

   return ret;
}

unsigned long MDrv_fsize (void *file)
{
   BUG_ON (file == NULL);
   return ((struct file *)file)->f_dentry->d_inode->i_size;
}

int MDrv_fread (void *file, unsigned long fpos, char *buf, unsigned long bufsize)
{
   BUG_ON (file == NULL);
   BUG_ON (buf == NULL);
   return kernel_read ((struct file *)file, fpos, buf, bufsize);
}

int MDrv_fclose (void *file)
{
   BUG_ON (file == NULL);
   return filp_close ((struct file *)file, 0);
}

/*
 * This function will read the virtual registry for a DWORD.
 * It will returns 0 on success
 */
int MDrv_read_dword_registry(const char* regEntry, MDrv_uint32 *Data)
{
   void * RegFile;
   void *Buffer = NULL;
   unsigned long fileSize;
   int status = 0;
   BUG_ON(regEntry == NULL);
   
   RegFile = MDrv_fopen (regEntry, MDRV_FRDONLY);
   if(RegFile == NULL)
   {
      return -1;
   }

   fileSize = MDrv_fsize(RegFile);

   if(fileSize > 0)
   {
      int readStatus;
      Buffer = MDrv_memalloc(fileSize);
      if(Buffer == NULL)
      {
         MDrv_fclose(RegFile);
         return -2;
      }

      readStatus = MDrv_fread (RegFile, 0, (char*)Buffer, fileSize);

      if(readStatus == fileSize)
      {
         MDrv_sscanf (Buffer, "%x", Data);
      }
      else
      {
         status = -3;
      }

      MDrv_memfree(Buffer);

   }
   else
      status = -4;

   MDrv_fclose(RegFile);

   return status;
}

unsigned long MDrv_getstate()
{
   unsigned long state = MDRV_USERCONTEXT;

   if (in_irq() || irqs_disabled())
      state = MDRV_IRQ;
   else if (in_softirq())
      state = MDRV_SIRQ;

   return state;
}

void MDrv_raisestate (unsigned long newstate, unsigned long *oldstate)
{
   unsigned long curstate = MDrv_getstate();
   BUG_ON (oldstate == NULL);

   if (newstate > curstate)
   {
      if (newstate > MDRV_SIRQ)
      {
         local_irq_save (*oldstate);
      }
      else if (newstate == MDRV_SIRQ)
      {
         local_bh_disable();
      }
   }
   else
      printk("useless %s\n", __FUNCTION__);
}

void MDrv_lowerstate (unsigned long newstate)
{
   unsigned long curstate = MDrv_getstate();

   if (curstate == MDRV_SIRQ)
   {
//      preempt_enable();
      local_bh_enable();
   }
   else
   {
//      nmi_exit();
      local_irq_restore (newstate);
   }
}

void *MDrv_map (unsigned long addr, unsigned long size, int NoErrorPrint)
{
   void *tmp = NULL;
   BUG_ON (in_interrupt());
   tmp = ioremap (addr, size);
   if (!NoErrorPrint && ((tmp == NULL) || IS_ERR(tmp)))
   {
      DbgCritical("failed for %lX size %lu: %p\n", addr, size, tmp);
      dump_stack();
      return NULL;
   }

   return tmp;
}

void *MDrv_map_cache (unsigned long addr, unsigned long size, int NoErrorPrint)
{
   void *tmp = NULL;
   BUG_ON (in_interrupt());
   tmp = ioremap_cache (addr, size);
   if (!NoErrorPrint && ((tmp == NULL) || IS_ERR(tmp)))
   {
      DbgCritical("failed for %lX size %lu: %p\n", addr, size, tmp);
      dump_stack();
      return NULL;
   }

   return tmp;
}

void *MDrv_map_wc (unsigned long addr, unsigned long size, int NoErrorPrint)
{
   void *tmp = NULL;
   BUG_ON (in_interrupt());
   tmp = ioremap_wc (addr, size);
   if (!NoErrorPrint && ((tmp == NULL) || IS_ERR(tmp)))
   {
      DbgCritical("failed for %lX size %lu: %p\n", addr, size, tmp);
      dump_stack();
      return NULL;
   }

   return tmp;
}


void MDrv_unmap (void *ptr)
{
   BUG_ON (in_interrupt());
   BUG_ON (ptr == NULL);

   iounmap (ptr);
}

void *MDrv_requestmem (unsigned long addr, unsigned long size, const char *ownername)
{
   BUG_ON (in_interrupt());
   return request_mem_region (addr, size, ownername);
}

void MDrv_releasemem (unsigned long addr, unsigned long size)
{
   BUG_ON (in_interrupt());
   release_mem_region (addr, size);
}

void MDrv_releaseresource (void *Resource)
{
   BUG_ON (in_interrupt());
   release_resource(Resource);
}

unsigned long MDrv_physaddr (void *virtaddr)
{
   return __pa(virtaddr);
}

void MDrv_gettime (MDrv_time_t *curtime)
{
   struct timeval tmp;
   BUG_ON (curtime == NULL);
   do_gettimeofday (&tmp);
   curtime->sec = tmp.tv_sec;
   curtime->usec = tmp.tv_usec;
}

void MDrv_wakeprocess (void *task)
{
   wake_up_process ((struct task_struct *)task);
}

MDrv_uint64 MDrv_getcounter64 (void)
{
   struct timespec ts;
   getrawmonotonic(&ts);
   return ((MDrv_uint64)ts.tv_sec) * NSEC_PER_SEC + ((MDrv_uint64)ts.tv_nsec);
}

int MDrv_getcounterfreq (void)
{
   return NSEC_PER_SEC;
}

MDrv_uint64 MDrv_getjiffies64 (void)
{
   return get_jiffies_64();
}

int MDrv_getjiffiesfreq (void)
{
   return HZ;
}

int MDrv_getcurrentprocid (void)
{
   return (current->tgid);
}

int MDrv_getcurrentprocgid (void)
{
   return (current->tgid);
}

void ospci2mdpci (void *os, MDrv_pcidev_t *md)
{
   md->irq     = ((struct pci_dev *)os)->irq;
   md->dev_id  = ((struct pci_dev *)os)->device;
   md->bus_nb  = ((struct pci_dev *)os)->bus->number;
   md->dev_fn  = ((struct pci_dev *)os)->devfn;
   md->os_dev  = (struct pci_dev *)os;
   md->dev_nb  = ((struct pci_dev *)os)->devfn >> 3;
   md->func_nb = ((struct pci_dev *)os)->devfn & 7;
}

/**
 * Returns the pci_dev struct pointer that matches the given DeviceId, VendorId and Index
 */
int MDrv_getpcidevicebyid (unsigned short DeviceId, unsigned short VendorId, unsigned short Index, MDrv_pcidev_t *pcidev)
{
   BUG_ON (pcidev == NULL);
   pcidev->os_dev = NULL;
   return MDrv_getnextpcidevicebyid (DeviceId, VendorId, Index, pcidev);
}

/**
 * Starting with the PCI device set in pcidev
 */
int MDrv_getnextpcidevicebyid (unsigned short DeviceId, unsigned short VendorId, unsigned short Index, MDrv_pcidev_t *pcidev)
{
   int i;
   int ret = -1;
   struct pci_dev *pci_device = NULL;
   unsigned int devId, venId;

   BUG_ON (pcidev == NULL);

   pci_device = (struct pci_dev *)pcidev->os_dev;

   /* We create a dummy devId and venId for the support of the PCI_ANY_ID */
   /* In the current implementation, pci_get_device takes unsigned int as arguments */
   /* and our function uses unsigned short. The PCI_ANY_ID is (~0), so in unsigned int */
   /* it will be 0xFFFFFFFF and in unsigned short, it will be 0xFFFF */
   /* In that case, the function will not do what we want */
   if(DeviceId == (unsigned short)MDRV_PCI_ANY_ID)
      devId = PCI_ANY_ID;
   else
      devId = DeviceId;

   if(VendorId == (unsigned short)MDRV_PCI_ANY_ID)
      venId = PCI_ANY_ID;
   else
      venId = VendorId;

   for (i = 0 ; i <= Index ; i++)
   {
      pci_device = pci_get_device(venId, devId, pci_device);

      if (pci_device == NULL)
         break;
   }
   if (pci_device)
   {
      ospci2mdpci(pci_device, pcidev);
      ret = 0;
   }

   return ret;
}

/**
 * Returns the pci_dev struct pointer that matches the given BusNb, DeviceNb, and FunctNb
 */
int MDrv_getpcidevicebynb(unsigned char BusNb, unsigned char DeviceNb, unsigned char FunctNb, MDrv_pcidev_t *pcidev)
{
   int func = 0;
   int ret = -1;
   int device = 0;
   struct pci_dev *pci_device = NULL;

   BUG_ON (pcidev == NULL);

   for_each_pci_dev(pci_device)
   {
      device = (unsigned char)(pci_device->devfn >> 3);
      func = (unsigned char)(pci_device->devfn & 7);
      if ((pci_device->bus->number) == BusNb && ((unsigned char)device) == DeviceNb && ((unsigned char)func) == FunctNb)
      {
         ospci2mdpci(pci_device, pcidev);
         ret = 0;
         break;
      }
   }

   return ret;
}

/**
 * Dereference the PCI device returned by a call to MDrv_getpcidevice...
 */
void MDrv_pci_dereference(MDrv_pcidev_t *pci_device)
{
   BUG_ON(pci_device == NULL);
   pci_dev_put(pci_device->os_dev);
}


int MDrv_pci_read_config_byte (MDrv_pcidev_t *pci_device, int where, MDrv_uint8 *Val)
{
   BUG_ON(pci_device == NULL);
   return pci_read_config_byte (pci_device->os_dev, where, Val);
}

int MDrv_pci_read_config_word (MDrv_pcidev_t *pci_device, int where, MDrv_uint16 *Val)
{
   BUG_ON(pci_device == NULL);
   return pci_read_config_word (pci_device->os_dev, where, Val);
}

int MDrv_pci_read_config_dword (MDrv_pcidev_t *pci_device, int where, MDrv_uint32 *Val)
{
   BUG_ON(pci_device == NULL);
   return pci_read_config_dword (pci_device->os_dev, where, Val);
}

int MDrv_pci_write_config_byte (MDrv_pcidev_t *pci_device, int where, MDrv_uint8 Val)
{
   BUG_ON(pci_device == NULL);
   return pci_write_config_byte (pci_device->os_dev, where, Val);
}

int MDrv_pci_write_config_word (MDrv_pcidev_t *pci_device, int where, MDrv_uint16 Val)
{
   BUG_ON(pci_device == NULL);
   return pci_write_config_word (pci_device->os_dev, where, Val);
}

int MDrv_pci_write_config_dword (MDrv_pcidev_t *pci_device, int where, MDrv_uint32 Val)
{
   BUG_ON(pci_device == NULL);
   return pci_write_config_dword (pci_device->os_dev, where, Val);
}

void MDrv_seminit (MDrv_sem_t *sem, int count)
{
   BUG_ON (sem == NULL);
   sem->os_sem = MDrv_memalloc (sizeof(struct semaphore));
   BUG_ON(sem->os_sem == NULL);
   sema_init ((struct semaphore *)sem->os_sem, count);
}

void MDrv_semdelete (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   MDrv_memfree (sem->os_sem);
}

void MDrv_mutexinit (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   sem->os_sem = MDrv_memalloc (sizeof(struct semaphore));
   BUG_ON(sem->os_sem == NULL);
   sema_init((struct semaphore *)sem->os_sem, 1);
}

void MDrv_semdown (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);
   if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
      DbgError("Wrong context\n");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)) || defined(CONFIG_PREEMPT_RT)
   down ((struct semaphore *)sem->os_sem);
#else
   down_killable ((struct semaphore *)sem->os_sem);
#endif
}

int MDrv_semdowninterruptible (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);
   if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
      DbgError("Wrong context\n");
   return down_interruptible ((struct semaphore *)sem->os_sem);
}

int MDrv_semtry (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);
   return down_trylock((struct semaphore *)sem->os_sem);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
void wait_for_sleeper(struct task_struct *arg)
{
   if (in_interrupt())
      printk("%s in interrupt\n", __FUNCTION__);
   set_tsk_thread_flag( arg, TIF_SIGPENDING );
   wake_up_process( arg );
}
#else
struct MDrv_delayed_work
{
   struct delayed_work worker;
   struct task_struct *arg;
};
void wait_for_sleeper(struct work_struct *work)
{
   struct MDrv_delayed_work *arg_struct = container_of(work, struct MDrv_delayed_work, worker.work);
   struct task_struct *thenCurrent = arg_struct->arg;
   if (in_interrupt())
      printk("%s in interrupt\n", __FUNCTION__);
   set_tsk_thread_flag(thenCurrent, TIF_SIGPENDING);
   wake_up_process(thenCurrent);
}
#endif
int MDrv_semdowntimeout (MDrv_sem_t *sem, unsigned long timeout)
{
   int ret = 0;
   unsigned long endtime = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
   struct work_struct work;
#else
   struct MDrv_delayed_work *work = MDrv_memalloc(sizeof(struct MDrv_delayed_work));
   work->arg = current;
#endif
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);

   if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
      DbgError("Wrong context\n");

   if (timeout == 0)
      return down_interruptible((struct semaphore *)sem->os_sem);

   endtime = timeout + jiffies;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
   INIT_WORK(&work, (void(*)(void*))wait_for_sleeper, (void*)(current));
   schedule_delayed_work(&work, timeout);
#else
   INIT_DELAYED_WORK(&work->worker, wait_for_sleeper);
   schedule_delayed_work(&work->worker, timeout);
#endif

   ret = down_interruptible((struct semaphore *)sem->os_sem);
   if (ret != 0)
   {
      while(time_before(jiffies, endtime))
      {
         ret = down_trylock((struct semaphore *)sem->os_sem);
         if (ret == 0)
            break;
         set_current_state(TASK_INTERRUPTIBLE);
         schedule_timeout(1);
      }
   }
   else
   {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
      cancel_delayed_work(&work);
#else
      cancel_delayed_work(&work->worker);
#endif
//      flush_scheduled_work();
      recalc_sigpending();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
      MDrv_memfree(work);
#endif
   }

   return ret;
}
#else // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
int MDrv_semdowntimeout (MDrv_sem_t *sem, unsigned long timeout)
{
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);

   if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
      DbgError("Wrong context\n");

   return down_timeout((struct semaphore *)sem->os_sem, timeout);
}
#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)

void MDrv_semup (MDrv_sem_t *sem)
{
   BUG_ON (sem == NULL);
   BUG_ON (sem->os_sem == NULL);
   up ((struct semaphore *)sem->os_sem);
}

void MDrv_spinlockinit (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   lock->os_spin = MDrv_memalloc (sizeof(spinlock_t));
   BUG_ON(lock->os_spin == NULL);
   spin_lock_init((spinlock_t *)lock->os_spin);
}

void MDrv_spinlockdelete (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   MDrv_memfree (lock->os_spin);
}

void MDrv_spinlock (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_lock((spinlock_t *)lock->os_spin);
}

void MDrv_spinunlock (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_unlock((spinlock_t *)lock->os_spin);
}

void MDrv_spinlock_bh (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_lock_bh((spinlock_t *)lock->os_spin);
}

void MDrv_spinunlock_bh (MDrv_spin_t *lock)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_unlock_bh((spinlock_t *)lock->os_spin);
}

void MDrv_spinlock_irqsave (MDrv_spin_t *lock, unsigned long *flags)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_lock_irqsave ((spinlock_t *)lock->os_spin, (*flags));
}

void MDrv_spinunlock_irqrestore (MDrv_spin_t *lock, unsigned long *flags)
{
   BUG_ON (lock == NULL);
   BUG_ON (lock->os_spin == NULL);
   spin_unlock_irqrestore ((spinlock_t *)lock->os_spin, (*flags));
}

int MDrv_requestirq (unsigned int irq, int (*handler)(int, void *), const char *devname, void *dev_id)
{
   return request_irq (irq, (irqreturn_t (*)(int, void*))handler, IRQF_SHARED, devname, dev_id);
}

void MDrv_freeirq (unsigned int irq, void *dev_id)
{
   free_irq (irq, dev_id);
}

void MDrv_synchronizeirq (unsigned int irq)
{
   synchronize_irq(irq);
}

void MDrv_disablepreempt (void)
{
   preempt_disable();
}

void MDrv_enablepreempt (void)
{
   preempt_enable();
}

void MDrv_inittasklet (MDrv_tasklet_t *task, void (*func)(unsigned long), unsigned long data)
{
   BUG_ON (task == NULL);
   task->os_tasklet = MDrv_memalloc (sizeof(struct tasklet_struct));
   BUG_ON(task->os_tasklet == NULL);
   tasklet_init (task->os_tasklet, func, data);
}

void MDrv_deletetasklet (MDrv_tasklet_t *task)
{
   BUG_ON (task == NULL);
   MDrv_memfree(task->os_tasklet);
}

int MDrv_scheduletasklet (MDrv_tasklet_t *task)
{
   int ret = 0;
   BUG_ON (task == NULL);
   BUG_ON (task->os_tasklet == NULL);

   ret = ((struct tasklet_struct *)task->os_tasklet)->state ^ (1 << TASKLET_STATE_SCHED);
   tasklet_schedule (task->os_tasklet);
   return ret;
}

void MDrv_canceltasklet(MDrv_tasklet_t *task)
{
   tasklet_kill(task->os_tasklet);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
struct MDrv_work_struct
{
   struct work_struct worker;
   void (*func)(unsigned long);
   unsigned long arg;
};
void MDrv_work_helper(struct work_struct *work)
{
   struct MDrv_work_struct *args = container_of(work, struct MDrv_work_struct, worker);
   args->func(args->arg);
}
#endif
void MDrv_init_work (MDrv_work_t *task, void (*func)(unsigned long), unsigned long data)
{
   BUG_ON (task == NULL);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
   task->os_work = MDrv_memalloc (sizeof(struct work_struct));
   BUG_ON(task->os_work == NULL);
   INIT_WORK ((struct work_struct *)(task->os_work), (void (*)(void *))func, (void *)data);
#else
   struct MDrv_work_struct *temp_work = MDrv_memalloc (sizeof(struct MDrv_work_struct));
   BUG_ON(temp_work == NULL);
   temp_work->func = func;
   temp_work->arg = data;
   INIT_WORK (&(temp_work->worker), MDrv_work_helper);
   task->os_work = &(temp_work->worker);
#endif
}

int MDrv_schedule_work (MDrv_work_t *task)
{
   BUG_ON (task->os_work == NULL);
   return schedule_work (task->os_work);
}

void *MDrv_startthread (int (*fn)(void *), void * arg)
{
   return kthread_run (fn, arg, "Matrox_kthread");
}

int MDrv_stopthread (void *thread)
{
   return kthread_stop((struct task_struct *)thread);
}

int MDrv_threadshouldstop (void)
{
   return kthread_should_stop();
}

unsigned int MDrv_readl(const void *addr)
{
   return readl(addr);
}

//////////////////////////////////////////////////////////////////////////////
//
// Name         : read_config_byte
//
// Synopsis     : Read a byte in the configuration registers
//
// Parameters   : struct pci_dev *dev: the device structure
//                int address: the address that will be read
//
// Return Value : the value that was read
//
//////////////////////////////////////////////////////////////////////////////
MDrv_uint8 read_config_byte (void *dev, int address)
{
   int result = 0;
   MDrv_uint8  reg_byte  = 0;

   BUG_ON (dev == NULL);

   result = pci_read_config_byte(dev, address, &reg_byte);
   return (result ? 0: reg_byte);
}

//////////////////////////////////////////////////////////////////////////////
//
// Name         : read_config_word
//
// Synopsis     : Read a 16 bits value in the configuration registers
//                dev->vendor
// Parameters   : struct pci_dev *dev: the device structure
//                int address: the address that will be read
//
// Return Value : the value that was read
//
//////////////////////////////////////////////////////////////////////////////
MDrv_uint16 read_config_word (void *dev, int address)
{
   int result = 0;
   MDrv_uint16 reg_word = 0;

   BUG_ON (dev == NULL);

   result = pci_read_config_word(dev, address, &reg_word);
   return (result ? 0: reg_word);
}

//////////////////////////////////////////////////////////////////////////////
//
// Name         : read_config_dword
//
// Synopsis     : Read a 32 bits in the configuration registers
//
// Parameters   : struct pci_dev *dev: the device structure
//                int address: the address that will be read
//
// Return Value : the value that was read
//
//////////////////////////////////////////////////////////////////////////////
MDrv_uint32 read_config_dword (void *dev, int address)
{
   int result = 0;
   MDrv_uint32  reg_dword  = 0;

   BUG_ON (dev == NULL);

   result = pci_read_config_dword(dev, address, &reg_dword);
   return (result ? 0: reg_dword);
}

int MDrv_delaythread (MDrv_uint64 interval, int alertable)
{
   // 1000*1000*10: conversion factor from seconds to hundredths of ns : Interval's time unit
   if (interval  < ((1000*1000*10) / MDrv_getcounterfreq()))
   {
      ndelay(interval*100);
   }
   else
   {
      wait_queue_head_t wait;
      if (in_interrupt() || in_atomic() || in_irq() || irqs_disabled())
         {
         DbgError("Wrong context\n");
         }

      init_waitqueue_head(&wait);
      interval /= (1000*1000*10) / MDrv_getcounterfreq() + 1;
      if (alertable)
      {
         if (wait_event_interruptible_timeout(wait, 0, interval)  == -ERESTARTSYS)
         {
            return -1;
         }
      }
      else
      {
         wait_event_timeout(wait, 0, interval);
      }
   }
   return 0;
}


void MDrv_msleep(unsigned int milisecond)
{
   msleep(milisecond);
}

void MDrv_usleep(unsigned long microsecond)
{
   udelay(microsecond);
}

void MDrv_nsleep(unsigned long nanosecond)
{
   ndelay(nanosecond);
}

unsigned long MDrv_pciresourcestart (MDrv_pcidev_t *dev, int bar)
{
   BUG_ON (dev == NULL);
   return pci_resource_start ((struct pci_dev *)dev->os_dev, bar);
}

unsigned long MDrv_pciresourcelen (MDrv_pcidev_t *dev, int bar)
{
   BUG_ON (dev == NULL);
   return pci_resource_len ((struct pci_dev *)dev->os_dev, bar);
}

int MDrv_pciresourceismem(MDrv_pcidev_t*dev, int bar)
{
   BUG_ON (dev == NULL);
   return pci_resource_flags ((struct pci_dev *)dev->os_dev, bar)  & IORESOURCE_MEM;
}

int MDrv_pciresourceisio(MDrv_pcidev_t*dev, int bar)
{
   BUG_ON (dev == NULL);
   return pci_resource_flags ((struct pci_dev *)dev->os_dev, bar)  & IORESOURCE_IO;
}

int MDrv_copytouser (unsigned char *user, unsigned char *kernel, int size)
{
   BUG_ON (user == NULL);
   BUG_ON (kernel == NULL);
   return copy_to_user (user, kernel, size);
}

int MDrv_copyfromuser (unsigned char *kernel, unsigned char *user, int size)
{
   BUG_ON (user == NULL);
   BUG_ON (kernel == NULL);
   return copy_from_user (kernel, user, size);
}

void MDrv_setinterruptible (void)
{
   set_current_state (TASK_INTERRUPTIBLE);
}

void MDrv_setrunning (void)
{
   set_current_state (TASK_RUNNING);
}

void MDrv_wmb (void)
{
   wmb();
}

void MDrv_schedule (void)
{
   set_current_state (TASK_INTERRUPTIBLE);
   schedule();
   set_current_state (TASK_RUNNING);
}

void *MDrv_getcurrenttask (void)
{
   return current;
}

void *MDrv_adddevice (long major, long offset, void *fops, const char *cdev_name, const char *device_name)
{
   struct cdev *dev_handle;

   if (! (dev_handle = cdev_alloc()) )
   {
      printk("Failed to allocate cdev\n");
   }
   else
   {
      dev_handle->owner = THIS_MODULE;
      kobject_set_name(&dev_handle->kobj, cdev_name, offset);
      dev_handle->ops = fops;
      if (cdev_add(dev_handle, major + offset, 1) < 0)
      { // Register one card3.
         printk("Failed to add device\n");
         kobject_put(&dev_handle->kobj); // remove object
         return 0;
      }
   }
   return dev_handle;
}

void MDrv_removedevice (void *dev_handle, long offset, const char *device_name)
{
   cdev_del (dev_handle);
}

void MDrv_flushsignals (void)
{
   flush_signals (current);
}

void MDrv_dumpstack (void)
{
   dump_stack ();
}

void MDrv_printsymbol (unsigned long addr)
{
   print_symbol ("%s\n", addr);
}

int MDrv_mtrr_add (void* addr, long size, int flags)
{
   return mtrr_add ((unsigned long)addr, size, flags, 1);
}

int MDrv_mtrr_del (int reg, void* addr, long size)
{
   return mtrr_del (reg, (unsigned long)addr, size);
}

void MDrv_initcompletion (MDrv_compl_t *comp)
{
   comp->os_compl = MDrv_memalloc (sizeof (struct completion));
   BUG_ON(comp->os_compl == NULL);
   init_completion (comp->os_compl);
}

void MDrv_reinitcompletion (MDrv_compl_t *comp)
{
   INIT_COMPLETION ((*(struct completion *)(comp->os_compl)));
}

void MDrv_deletecompletion (MDrv_compl_t *comp)
{
   MDrv_memfree (comp->os_compl);
}

void MDrv_waitcompletion (MDrv_compl_t *comp)
{
   wait_for_completion (comp->os_compl);
}

int MDrv_waitcompletion_interruptible (MDrv_compl_t *comp)
{
   return wait_for_completion_interruptible (comp->os_compl);
}

unsigned long MDrv_waitcompletion_timeout (MDrv_compl_t *comp, unsigned  long timeout)
{
   return wait_for_completion_timeout (comp->os_compl, timeout);
}

unsigned long MDrv_waitcompletion_interruptible_timeout (MDrv_compl_t *comp, unsigned  long timeout)
{
   return wait_for_completion_interruptible_timeout (comp->os_compl, timeout);
}

void MDrv_signalcompletion (MDrv_compl_t *comp)
{
   complete_all (comp->os_compl);
}

int MDrv_alloc_uart_port (MDrv_uart_t *uart)
{
   BUG_ON (uart == NULL);
   uart->os_uart = MDrv_memalloc (sizeof (struct uart_port));
   BUG_ON(uart->os_uart == NULL);
   if (uart)
      return 0;
   else
      return -ENOMEM;
}

void MDrv_free_uart_port (MDrv_uart_t *uart)
{
   BUG_ON (uart == NULL);
   kfree (uart);
}

unsigned char MDrv_inb(unsigned short port)
{
   return inb(port);
}

void MDrv_outb(unsigned char value, unsigned short port)
{
   return outb(value, port);
}

unsigned short MDrv_inw(unsigned short port)
{
   return inw(port);
}

void MDrv_outw(unsigned short value, unsigned short port)
{
   return outw(value, port);
}

unsigned int MDrv_inl(unsigned short port)
{
   return inl(port);
}

void MDrv_outl(unsigned int value, unsigned short port)
{
   return outl(value, port);
}
