#!/bin/sh

######################################################################
#
# Filename: mil-uninstall.sh
#
# Content: Uninstalls every MIL packages.
#          Must be run as root.
#
# Copyright (c) 2007 Matrox Electronic Systems Ltd.
# All Right Reserved
#
######################################################################

ROOT_UID=0
LOG_FILE="/var/log/mil-installer-32.log"
RPM_PACKAGE=1
DEB_PACKAGE=0
USED_PACKAGE=$RPM_PACKAGE
ROOT_ERROR_CODE=1

if uname -m|grep x86_64 2>&1 >/dev/null; then
LOG_FILE="/var/log/mil-installer-64.log"
fi

outputLog()
{
    echo "[MIL     ]"`LC_ALL=C date "+%a %b %d %H:%M:%S %Y"`" - $1" >> $LOG_FILE
}

outputLogError()
{
    echo "[MIL     ]"`LC_ALL=C date "+%a %b %d %H:%M:%S %Y"`" ---- $1" >> $LOG_FILE
}

outputConsole()
{
    echo "$1"
}

outputBoth()
{
    outputLog "$1"
    outputConsole "$1"
}

uninstallPackage()
{
    if test "$USED_PACKAGE" = "$RPM_PACKAGE" ; then
	if rpm -q "$1" >/dev/null 2>&1
	then
	    if output=$(rpm -e "$1" 2>&1); then
		outputLog "Package $1 uninstalled"
	    else
		outputLogError "$output"
		outputConsole "$output"
	    fi
	fi
    else
	if output=$(dpkg --purge "$1" 2>&1); then
	    outputLog "Package $1 uninstalled"
	else
	    outputLogError "$output"
	    outputConsole "$output"
	fi
    fi
}

if test -z "$UID"; then
    UID=$(id -u)
fi

if test "$UID" != "$ROOT_UID" ; then
   outputConsole "The uninstall script must be run as root!"
   exit $ROOT_ERROR_CODE
fi

outputLog "[Initialize]"
outputLog "[Start of uninstall]"

outputBoth "MIL uninstall in progress, please wait..."

if [ -f /etc/debian_version ]; then
    USED_PACKAGE=$DEB_PACKAGE
fi

set -- mil-full-drivers mil-full-examples mil-full-dev mil-full-display mil-full
set -- "$@" mil-lite-drivers mil-lite-examples mil-lite-dev mil-lite-display mil-lite

for package do
    outputLog "Uninstalling $package"
    uninstallPackage $package
done

outputLog "[End of uninstall]"
outputConsole "Uninstallation complete!"

exit 0
